@@ -29,12 +29,14 @@ KEY_VOCABULARIES = {
 }
 
 LANG_REGEX = re.compile(r"^lang(?::(.+))?")
+AGGREGATES = ['sample', 'count', 'sum', 'min', 'max', 'avg']
 
 logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.WARNING)
 logger = logging.getLogger('sparql_transformer')
 
 
-def pre_process(_input, options=None):
+def pre_process(json_query, options=None):
+    _input = json_query.copy()
     opt = DEFAULT_OPTIONS.copy()
     if '@context' in _input:
         opt['context'] = _input['@context']
@@ -72,6 +74,10 @@ def pre_process(_input, options=None):
     opt['voc'] = voc
     opt['is_json_ld'] = is_json_ld
 
+    if '$limitMode' in json_query and '$limit' in json_query:
+        opt['limit'] = json_query['$limit']
+        opt['offset'] = json_query.get('$offset', 0)
+
     return query, proto, opt
 
 
@@ -83,20 +89,26 @@ def post_process(sparql_res, proto, opt):
     instances = list(map(lambda b: _sparql2proto(b, proto, opt), bindings))
     # merge lines with the same id
     content = []
-    anchor = instances[0]['$anchor'] if len(instances) > 0 else 'id'
-    for inst in instances:
-        _id = inst[anchor]
-        # search if we have already the same id
-        match = [x for x in content if x[anchor] == _id]
-        if not match:  # it is a new one
-            content.append(inst)
-        else:  # otherwise modify previous one
-            _merge_obj(match[0], inst)
+    anchor = instances[0]['$anchor'] if (len(instances) > 0 and '$anchor' in instances[0]) else None
+    if not anchor:
+        content = instances
+    else:
+        for inst in instances:
+            _id = inst[anchor]
+            # search if we have already the same id
+            match = [x for x in content if x[anchor] == _id]
+            if not match:  # it is a new one
+                content.append(inst)
+            else:  # otherwise modify previous one
+                _merge_obj(match[0], inst)
 
     # remove anchor tag
     for i in content:
         clean_recursively(i)
 
+    if 'limit' in opt:
+        content = content[opt['offset']: opt['offset'] + opt['limit']]
+
     if is_json_ld:
         return {
             '@context': opt['context'],
@@ -107,7 +119,6 @@ def post_process(sparql_res, proto, opt):
 
 def sparqlTransformer(_input, options=None):
     query, proto, opt = pre_process(_input, options)
-
     sparql_fun = opt['sparqlFunction'] if 'sparqlFunction' in opt else _default_sparql(opt['endpoint'])
     sparql_res = sparql_fun(query)
 
@@ -143,8 +154,10 @@ def _jsonld2query(_input):
     wheres = [w for w in wheres if w]
 
     _from = ('FROM <%s>' % modifiers['$from']) if '$from' in modifiers else ''
-    limit = ('LIMIT %d' % modifiers['$limit']) if '$limit' in modifiers else ''
-    offset = 'OFFSET ' + modifiers['$offset'] if '$offset' in modifiers else ''
+    limit = ('LIMIT %d' % modifiers['$limit']) if (
+            '$limit' in modifiers and modifiers.get('$limitMode') != 'library') else ''
+    offset = 'OFFSET ' + str(modifiers['$offset']) if (
+            '$offset' in modifiers and modifiers.get('$limitMode') != 'library') else ''
     distinct = '' if ('$distinct' in modifiers and modifiers['$distinct'] == 'false') else 'DISTINCT'
     prefixes = _parse_prefixes(modifiers['$prefixes']) if '$prefixes' in modifiers else []
     values = parse_values(values_normalized) if '$values' in modifiers else []
@@ -242,11 +255,14 @@ def _fit_in(instance, line, options):
 
         # if value is an obj
         if isinstance(variable, dict):
+            obj_as_list = variable.get('$list', False)
             fii_fun = _fit_in(variable, line, options)
             for key in list(variable):
                 fii_fun(key)
             if _is_empty_obj(variable):
                 instance.pop(k)
+            elif obj_as_list:
+                instance[k] = [instance[k]]
             return
 
         if not isinstance(variable, str):
@@ -258,6 +274,9 @@ def _fit_in(instance, line, options):
         variable = variable[1:]
         accept = None
         langTag = options['langTag']
+        asList = "$list" in variable
+        variable = variable.replace("$list", "")
+
         if "$accept:" in variable:
             temp = variable.split('$accept:')
             variable = temp[0]
@@ -274,6 +293,7 @@ def _fit_in(instance, line, options):
             opt = options.copy()
             opt['accept'] = accept
             opt['langTag'] = langTag
+            opt['list'] = asList
             instance[k] = _to_jsonld_value(line[variable], opt)
 
             if instance[k] is None:
@@ -335,7 +355,7 @@ def _to_jsonld_value(_input, options):
 
     # nothing more to do for other types
     if not isinstance(value, str):
-        return value
+        return [value] if options['list'] else value
 
     # if here, it is a string or a date, that are not parsed
     if 'xml:lang' in _input and options['langTag'] != 'hide':
@@ -347,7 +367,7 @@ def _to_jsonld_value(_input, options):
                 voc['lang']: lang,
                 voc['value']: value
             }
-    return value
+    return [value] if options['list'] else value
 
 
 def _merge_obj(base, addition):
@@ -368,6 +388,10 @@ def _merge_obj(base, addition):
         if isinstance(a, dict) and '$anchor' in a:
             anchor = a['$anchor']
 
+        # if a is array, I take its value
+        if isinstance(a, list):
+            a = a[0]
+
         if isinstance(b, list):
             if anchor:
                 same_ids = [x for x in b if anchor in x and a[anchor] == x[anchor]]
@@ -423,6 +447,7 @@ def _compute_root_id(proto, prefix):
         proto[k] += '$var:' + _rootId
 
     proto['$anchor'] = k
+    proto['$list'] = '$list' in proto[k]
     return _rootId, required
 
 
@@ -437,7 +462,7 @@ def _manage_proto_key(proto, vars=[], filters=[], wheres=[], main_lang=None, pre
     _rootId = _rootId or prev_root or '?id'
 
     def inner(k, i=''):
-        if [k] == '$anchor':
+        if k == '$anchor' or k == '$list':
             return
         v = proto[k]
         if isinstance(v, dict):
@@ -466,24 +491,37 @@ def _manage_proto_key(proto, vars=[], filters=[], wheres=[], main_lang=None, pre
             options = v.split('$')
             v = options.pop(0)
 
-        id = ('?' + prefix + str(i)) if is_dollar else v
+        original_id = ('?' + prefix + str(i)) if is_dollar else v
+        id = original_id
+
         _var = [s for s in options if s.startswith('var:')]
         if len(_var) > 0:
             id = _sparql_var(_var[0].split(':')[1])
             if not id.startswith('?'):
                 id = '?' + id
 
-        required = 'required' in options or k in ['id', '@id'] or id in values
+        _accept = [s for s in options if s.startswith('accept')]
+        _bestlang = [s for s in options if s.startswith('bestlang')]
+        _langTag = [s for s in options if s.startswith('langTag')]
+
+        aggregate = [a for a in AGGREGATES if a in options]
+        aggr_what = id if is_dollar else original_id
+        if len(aggregate) > 0 and len(_var) == 0:
+            id = original_id if is_dollar else f"?{aggregate[0]}_{original_id.replace('?', '')}"
+
+        required = 'required' in options or k in ['id', '@id'] or id in values or (len(aggregate) > 0 and is_dollar)
         # if it is an id or I specified a value, this property can not be optional
 
         proto[k] = id
-        _bestlang = [s for s in options if s.startswith('bestlang')]
-        _langTag = [s for s in options if s.startswith('langTag')]
 
         _var = id
         if 'sample' in options:
             _var = '(SAMPLE(%s) AS %s)' % (id, id)
 
+        if len(aggregate) > 0:
+            distinct_txt = 'DISTINCT ' if 'distinct' in options else ''
+            _var = f"({aggregate[0].upper()}({distinct_txt}{aggr_what}) AS {id})"
+
         if len(_bestlang) > 0:
             _bestlang = _bestlang[0]
             proto[k] = id + '$accept:string'
@@ -492,11 +530,17 @@ def _manage_proto_key(proto, vars=[], filters=[], wheres=[], main_lang=None, pre
                 raise AttributeError('bestlang require a language declared inline or in the root')
 
             _var = '(sql:BEST_LANGMATCH(%s, "%s", "en") AS %s)' % (id, lng, id)
+        elif len(_accept) > 0:
+            proto[k] = id + '$' + _accept[0]
 
         if len(_langTag) > 0:
             proto[k] = proto[k] + '$' + _langTag[0]
 
-        vars.append(_var)
+        if 'list' in options and id != _rootId:
+            proto[k] += '$list'
+
+        if _var not in vars:
+            vars.append(_var)
 
         # lang filters are managed here, so that they stay within the OPTIONAL
         lang_filter = ''
@@ -549,6 +593,7 @@ def clean_recursively(instance):
 
     if isinstance(instance, dict):
         instance.pop('$anchor', None)  # remove $anchor
+        instance.pop('$list', None)  # remove $anchor
         for k, v in instance.items():
             clean_recursively(v)
 
